#include "main.h"

struct globals_t globals;

void sig_proc(int sig)
{
	/* AUDIT: None of this code is signal safe, it does a hell of
	 * a lot of stuff with file descriptrors. */
	switch (sig){
		case SIGHUP:
			if ( flag(FLAG_STDOUT) || flag(FLAG_LOGFILE) ){
				print_close();
				if (!print_init()) {
					print_err("ERROR: Couldn't reopen "
					"print/logfile descriptors" CRLF);
					exit(0);
				}
			}

			if ( flag(FLAG_SYSLOG)  ) {
				/* Can't really fail */
				slog_close();
				slog_init();
			}
			
#ifdef _HAVE_LIBPCAP
			if ( flag(FLAG_LIBPCAP) ) {
				lpcap_close();
				if (!lpcap_init()) {
					print_err("ERROR: Couldn't reopen "
					"tcpdump file" CRLF);
					print_close();
					exit(0);
				}
			}
#endif /* _HAVE_LIBPCAP */

			break;
		default:
			return;
	}

	return;
}

/* Decide what to do with a packet */
void dump_packet ( p_packet p )
{
	time_t now;

	/* Attach the time to this packet */
	if ( time(&now) ) p->time=localtime(&now);

	/* Perform basic demultiplexing */
	p->ip = (struct iphdr *)&p->data.data;
	p->trans = (p->data.data) + (p->ip->ihl*4);
	p->app = NULL;

	/* AUDIT: Stupid behaivour */
	/* if ( p->ip->protocol == IPPROTO_IGMP ) return; */

	if ( (flag(FLAG_STDOUT) && !flag(FLAG_DAEMON) ) || flag(FLAG_LOGFILE) ) {
		print_dump(p);
	}

	if ( flag(FLAG_SYSLOG) ) {
		slog_dump(p);
	}
	
#ifdef _HAVE_LIBPCAP
	if ( flag(FLAG_LIBPCAP) ) {
		lpcap_dump(p);
	}
#endif
}

/* Display usage information */
void usage(char *apppath)
{
	if (apppath==NULL) apppath="";

	print_err("Usage:" CRLF);
	print_err("  %s [OPTIONS]" CRLF CRLF, apppath);

	/* Output */
	print_err("  -t <filename>\tName of file to output tcpdump logs" CRLF);
	print_err("  -l <filename>\tName of file to log to" CRLF);
	print_err("  -s \t\tSilent, no output to screen" CRLF);
	print_err("  -d \t\tDaemon mode, silent, and backgrounded" CRLF);
	print_err("  -y \t\tLog machine parsable output to syslog" CRLF);

	/* Formatting */
	print_err("  -n <length>\tLine length for hex/ascii output" CRLF);
	print_err("  -a \t\tShow the date as well as time on packets" CRLF);
	print_err("  -m \t\tDo not dump data" CRLF);
	print_err("  -x \t\tDo not display hex output" CRLF);
	print_err("  -i \t\tDo not display ascii output" CRLF);

	/* Security */
	print_err("  -u <uid>\tUser ID to drop to" CRLF);
	print_err("  -g <gid>\tGroup ID to drop to" CRLF);
	print_err("  -c <chroot>\tDirectory to chroot jail inside" CRLF);

	/* Help */
	print_err("  -v \t\tJust print version" CRLF);
	print_err("  -h \t\tDisplay this help" CRLF CRLF);

	if ( err(ERROR_SUGID) ){
		print_err("%s: " ERR_SUGID CRLF, apppath);
	}
	if ( err(ERROR_FMATCH) ){
		print_err("%s: " ERR_FMATCH CRLF, apppath);
	}
	if ( err(ERROR_NODATA) ){
		print_err("%s: " ERR_NODATA CRLF, apppath);
	}
	if ( err(ERROR_ZEROLINE) ){
		print_err("%s: " ERR_ZEROLINE CRLF, apppath);
	}	
	if ( err(ERROR_LIBPCAP) ){
		print_err("%s: " ERR_LIBPCAP CRLF, apppath);
	}
	
}

/* Parse the command line for options */
int parse_command_line( int argc, char **argv )
{
	int i, ret=1;

	while ( (i=getopt(argc,argv, "t:l:sdhvu:g:c:axmin:y" ))!=-1){
    	switch (i)
    	{
		case 'y':
			flag_set(FLAG_SYSLOG);
			continue;
    		case 'n':
    			globals.length = atoi(optarg);
    			continue;
    		case 'i':
    			flag_set(FLAG_NOASCII);
    			continue;
    		case 'x':
    			flag_set(FLAG_NOHEX);
    			continue;
    		case 'm':
    			flag_set(FLAG_NODATA);
    			continue;
    		case 'a':
    			flag_set(FLAG_SHOWDATE);
    			continue;
    		case 'c':
    			flag_set(FLAG_CHROOT);
    			globals.chroot = optarg;
    			continue;
    		case 'u':
    			flag_set(FLAG_SUID);
    			globals.uid = atoi(optarg);
    			continue;
    		case 'g':
    			flag_set(FLAG_SGID);
    			globals.gid = atoi(optarg);
    			continue;
    		case 't':
#ifdef _HAVE_LIBPCAP
    			flag_set(FLAG_LIBPCAP);
    			globals.pcapfile=optarg;
#else
			err_set(ERROR_LIBPCAP);
			ret=0;
#endif
			continue;
    		case 'l':
    			flag_set(FLAG_LOGFILE);
    			globals.logfile=optarg;
    			continue;
    		case 's':
 			flag_unset(FLAG_STDOUT);
    			continue;
    		case 'd':
    			flag_unset(FLAG_STDOUT);
    			flag_set(FLAG_DAEMON);
    			continue;
    		case 'v':
    			print_err("%s: " VERSION_STRING CRLF CRLF, argv[0]);
    			exit(0);
    		case 'h':
    		case '?':
    		default:
    			return 0;
    	}
    }

	
	if ( (flag(FLAG_SUID) || flag(FLAG_SGID)) && 
			!(flag(FLAG_SUID) && flag(FLAG_SGID)) ) {
		err_set(ERROR_SUGID);
		ret = 0;
	}

	if ( flag(FLAG_LIBPCAP) && flag(FLAG_LOGFILE) ) {
		if ( !strcmp(globals.logfile, globals.pcapfile) ) {
			err_set(ERROR_FMATCH);
			ret = 0;
		}
	}
	
	/* AUDIT: length must be 1 or more,
	 * not 0 or more, otherwise hex dump
	 * loops infinately */
	if ( globals.length <= 0 ) {
		err_set(ERROR_ZEROLINE);
		ret = 0;
	}

	if ( flag(FLAG_NOASCII) && flag(FLAG_NOHEX) ) {
		err_set(ERROR_NODATA);
		ret = 0;
	}
	
	return ret;
}

/* DAEMONIZE - drops root privs, and chroots etc.. */
int daemonize(void)
{
	if ( flag(FLAG_CHROOT) ) {
		if ( chdir(globals.chroot) == -1 ) {
			perror("chdir()");
			return 0;
		}
		if ( chroot(".") == -1 ) {
			perror("chroot()");
			return 0;
		}
	}

	if ( flag(FLAG_SGID) ) {
		if (setgroups(1,&globals.gid) == -1) {
			perror("setgroups()");
			return 0;
		}
		
		/* AUDIT: Check we actually dropped GID! */
		if ( setgid(globals.gid) == -1 ) {
			perror("setgid()");
			return 0;
		}
	}

	if ( flag(FLAG_SUID) ){
		/* AUDIT: Check we actually dropped UID! */
		if ( setuid(globals.uid) == -1 ) {
			perror("setuid()");
			return 0;
		}
	}

	/* background our process */
	if ( flag(FLAG_DAEMON) ) {
		switch(fork())
		{
			case 0: break;
			case -1:
				perror("fork()");
				return 0;
			default: _exit(0);
		}

		if (setsid() < 0) {
			perror("setsid()");
			return 0;
		}
		
		switch(fork())
		{
			case 0: break;
			case -1:
				perror("fork() again");
				return 0;
			default: _exit(0);
		}
	}

	return 1;
}


/* APP ENTRY POINT */
int main ( int argc, char **argv )
{
	int len;
	int fd=0, fds=sysconf(_SC_OPEN_MAX);
	packet_t packet;

	/* zero global data */
	memset(&globals, 0, sizeof(globals));

	/* Set default flags */
	flag_set(FLAG_STDOUT);
	globals.length = LINE_LENGTH;
	
	/* Set timezone */
	tzset();

	if ( !parse_command_line(argc, argv) ) {
		usage(argv[0]);
		exit(1);
	}

#ifdef _DEBUG
	if ( flag(FLAG_LIBPCAP) ) print_err("TCPDUMP file: %s" CRLF, globals.pcapfile);
	if ( flag(FLAG_LOGFILE) ) print_err("LOGFILE file: %s" CRLF, globals.logfile);
	if ( flag(FLAG_SYSLOG) ) print_err("SYSLOG" CRLF);
	if ( flag(FLAG_CHROOT) ) print_err("CHROOT: %s" CRLF, globals.chroot);
	if ( flag(FLAG_SUID) ) print_err("SUID: %i" CRLF, globals.uid);
	if ( flag(FLAG_SGID) ) print_err("SGID: %i" CRLF, globals.gid);
	if ( flag(FLAG_DAEMON) )  print_err("DAEMON MODE" CRLF);
	if ( !flag(FLAG_STDOUT) )  print_err("SILENT MODE" CRLF);
#endif

	/* Need this _before_ we open netlink/output files ;) */
	if ( flag(FLAG_DAEMON) ) {
		while (fd<fds) close(fd++);
	}

	if ( !netlink_open() ) {
		print_err(ERR_OPEN_NETLINK CRLF);
		exit (1);
	}

	if ( !daemonize() ) {
		print_err("Error daemonising" CRLF);
	}
	
	/* Initialise various output mechanisms
	 * this opens file descriptors */
	if (!print_init()) {
		print_err("print_init()" CRLF);
		return 1;
	}

	if (!slog_init()) {
		print_err("slog_init()" CRLF);
		return 1;
	}
	
#ifdef _HAVE_LIBPCAP
	if (!lpcap_init()) {
		print_err("lpcap_init()" CRLF);
		print_close();
		return 1;
	}
#endif

	if (flag(FLAG_DAEMON)) {
		signal(SIGHUP, sig_proc);
	}

	/* Enter the main program loop */
	while (1){
		len = netlink_read( &packet.data, sizeof(packet.data));

		if ( len <= 0 ) {
			if ( !flag(FLAG_DROPROOT) ) {
				/* eek netlink died, reopen it */
				if ( len < 0 ) perror("read()");
				sleep(1);
				netlink_close();
				netlink_open(); /* best effort */
				continue;
			}else{
				print_err(ERR_RESYNC_NOROOT CRLF);
				break;
			}
		}

		if ( len < (sizeof(packet.data) - BUFFER_SIZE) ){
			print_err(ERR_PACKET_TOO_SMALL CRLF);
		}else{
			dump_packet(&packet);
		}
	}

	netlink_close();
	exit(1);
}
